<h3>Download</h3>

<p>Current version: <b>{{ version }}</b></p>
<p>Get Setuptools from the <a href="https://pypi.python.org/pypi/setuptools"> Python Package Index</a>

<h3>Questions? Suggestions? Contributions?</h3>

<p>Visit the <a href="https://bitbucket.org/pypa/setuptools">Setuptools project page</a> </p>
